# Toolchain builder for the Serene programming language
#
# Copyright (c) 2019-2023 Sameer Rahmani <lxsameer@gnu.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os

from conan import ConanFile
from conan.tools.cmake import CMake, CMakeToolchain, CMakeDeps
from conan.tools.files import get, copy

from conf.utils import get_version, with_musl_toolchain


class ZlibBase(ConanFile):
    version = get_version("zlib")
    settings = "os", "arch", "build_type", "compiler"

    exports = "*.cmake"
    cmake_setup_file = "zlib_setup.cmake"

    def source(self):
        get(
            self,
            f"https://github.com/madler/zlib/archive/v{self.version}.tar.gz",
        )

    def generate(self):
        tc = CMakeToolchain(self)
        tc.variables["BUILD_SHARED_LIBS"] = False
        with_musl_toolchain(self, tc)
        tc.generate()
        deps = CMakeDeps(self)
        deps.generate()

    def build(self):
        cmake = CMake(self)
        cmake.configure(
            {
                "BUILD_SHARED_LIBS": False,
                "CMAKE_POSITION_INDEPENDENT_CODE": "ON",
            },
            build_script_folder=f"zlib-{self.version}",
        )
        self.run("ninja zlibstatic")

    def package(self):
        copy(self, "*.a", self.build_folder, os.path.join(self.package_folder, "lib"))
        copy(self, self.cmake_setup_file, self.recipe_folder, self.package_folder)
        copy(
            self,
            "zconf.h",
            self.build_folder,
            os.path.join(self.package_folder, "include"),
        )
        copy(
            self,
            "zlib.h",
            os.path.join(self.build_folder, f"zlib-{self.version}"),
            os.path.join(self.package_folder, "include"),
        )
        copy(
            self,
            "zlib.pc",
            self.build_folder,
            os.path.join(self.package_folder, "share", "pkgconfig"),
        )

    def package_info(self):
        self.cpp_info.set_property("cmake_build_modules", [self.cmake_setup_file])
        self.cpp_info.set_property("cmake_find_mode", "module")
        self.cpp_info.set_property("cmake_file_name", "ZLIB")
        self.cpp_info.set_property("cmake_target_name", "ZLIB::ZLIB")
        self.cpp_info.set_property("pkg_config_name", "zlib")
        self.cpp_info.libs = ["z"]
        self.cpp_info.set_property("cmake_find_package", "ZLIB")
        self.cpp_info.set_property("cmake_find_package_multi", "ZLIB")
