# Toolchain builder for the Serene programming language
#
# Copyright (c) 2019-2023 Sameer Rahmani <lxsameer@gnu.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
from pathlib import Path

from conan import ConanFile
from conan.tools.cmake import CMake, CMakeToolchain, CMakeDeps

from conan.tools.files import get

from conf.utils import with_static_flags, current_dir, get_version

PROJECTS = [
    "clang",
    "lld",
]

TOOLCHAIN_TARGETS = [
    "clang-resource-headers",
    "clang",
    "lld",
    "llvm-ar",
    "llvm-config",
    "llvm-nm",
    "llvm-objcopy",
    "llvm-profdata",
    "llvm-ranlib",
    "llvm-strip",
    "llvm-tblgen",
    "llvm-strip",
    "llvm-readelf",
    "llvm-windres",
    "LTO",
    "llvm-lipo",
    "runtimes",
    "builtins",
]

RUNTIME_TARGETS = [
    "compiler-rt",
    "libcxx",
    "libcxxabi",
    "libunwind",
]

CMAKE_OPTIONS = {
    "CLANG_DEFAULT_CXX_STDLIB": "libc++",
    "CLANG_DEFAULT_OBJCOPY": "llvm-objcopy",
    "CLANG_DEFAULT_RTLIB": "compiler-rt",
    "CLANG_DEFAULT_UNWINDLIB": "libunwind",
    "CLANG_DEFAULT_LINKER": "lld",
    "LLVM_USE_LINKER": "lld",
    "CLANG_VENDOR_UTI": "serene.stage0",
    "LLVM_ENABLE_NEW_PASS_MANAGER": "ON",
    "LLVM_BUILD_TESTS": "OFF",
    "LLVM_ENABLE_ASSERTIONS": "OFF",
    "LLVM_ENABLE_LIBXML2": "OFF",
    "LLVM_ENABLE_TERMINFO": "OFF",
    "LLVM_ENABLE_ZLIB": "FORCE_ON",
    "LLVM_INCLUDE_BENCHMARKS": "OFF",
    "LLVM_INCLUDE_EXAMPLES": "OFF",
    "LLVM_INCLUDE_TESTS": "OFF",
    "LLVM_INCLUDE_GO_TESTS": "OFF",
    "LLVM_ENABLE_BINDINGS": "OFF",
    "LLVM_TARGETS_TO_BUILD": "X86",
    "LLVM_STATIC_LINK_CXX_STDLIB": "ON",
    "PACKAGE_VENDOR": "serene.stage0",
    "LLVM_ENABLE_PER_TARGET_RUNTIME_DIR": "OFF",
    "ENABLE_X86_RELAX_RELOCATIONS": "ON",
    "BUILD_SHARED_LIBS": "OFF",
    "CLANG_ENABLE_BOOTSTRAP": "ON",
    # "LLVM_DEFAULT_TARGET_TRIPLE": "x86_64-pc-linux-musl",
    "LIBCXX_ENABLE_STATIC_ABI_LIBRARY": "ON",
    "LIBCXXABI_ENABLE_SHARED": "OFF",
    "LIBCXX_ABI_VERSION": "2",
    "LLVM_CCACHE_BUILD": "ON",
    "LLVM_ENABLE_LTO": "OFF",
    # Common
    "CMAKE_POSITION_INDEPENDENT_CODE": "ON",
}


class ClangBootstrap(ConanFile):
    name = "clang-bootstrap"
    settings = "os", "arch", "build_type"
    version = get_version("llvm")

    gnu_triple = "x86_64-build-linux-gnu"
    musl_triple = os.environ["TARGET"]

    def build_requirements(self):
        self.requires(f"llvm-source/{self.version}@{self.user}/{self.channel}")

        self.requires(
            f"zlib-bootstrap/{get_version('zlib')}@{self.user}/{self.channel}",
            visible=False,
            transitive_libs=False,
        )
        self.requires(
            f"zstd-bootstrap/{get_version('zstd')}@{self.user}/{self.channel}",
            visible=False,
            transitive_libs=False,
        )
        self.tool_requires(f"cmake/{get_version('cmake')}@{self.user}/{self.channel}")
        self.tool_requires(f"ninja/{get_version('ninja')}@{self.user}/{self.channel}")

    def generate(self):
        tc = CMakeToolchain(self)
        tc.generate()

        deps = CMakeDeps(self)
        deps.set_property("zlib", "cmake_find_mode", "both")
        deps.generate()

    def build(self):
        envs = self.buildenv_info.vars(self, scope="build")
        cflags = envs.get("CFLAGS", "")
        cxxflags = envs.get("CXXFLAGS", "")
        ldflags = envs.get("LDFLAGS", "")
        asflags = envs.get("ASFLAGS", "")

        targets = [self.musl_triple, self.gnu_triple]

        opts = CMAKE_OPTIONS
        opts["LLVM_ENABLE_PROJECTS"] = ";".join(PROJECTS)
        opts["LLVM_ENABLE_RUNTIMES"] = ";".join(RUNTIME_TARGETS)
        opts["LLVM_BUILTIN_TARGETS"] = ";".join(targets)
        opts["LLVM_RUNTIME_TARGETS"] = ";".join(targets)
        opts["LLVM_DISTRIBUTION_COMPONENTS"] = ";".join(TOOLCHAIN_TARGETS)

        # opts["LLVM_RUNTIME_DISTRIBUTION_COMPONENTS"] = "cxx-headers"

        zlibdir = self.dependencies["zlib-bootstrap"]
        zlib_include = zlibdir.cpp_info.includedir
        zlib_lib = zlibdir.cpp_info.libdir
        opts["CMAKE_REQUIRED_INCLUDES"] = zlib_include
        opts["CMAKE_REQUIRED_LIBRARIES"] = f"-L{zlib_lib} -lz"

        for target in targets:
            opts[f"BUILTINS_{target}_CMAKE_SYSTEM_NAME"] = "Linux"
            opts[f"BUILTINS_{target}_CMAKE_BUILD_TYPE"] = "Release"
            opts[f"BUILTINS_{target}_CMAKE_SYSTEM_PROCESSOR"] = "x86_64"
            opts[f"BUILTINS_{target}_CMAKE_C_FLAGS"] = cflags
            opts[f"BUILTINS_{target}_CMAKE_CXX_FLAGS"] = cxxflags
            opts[f"BUILTINS_{target}_CMAKE_ASM_FLAGS"] = asflags
            opts[f"BUILTINS_{target}_CMAKE_C_COMPILER_TARGET"] = target
            opts[f"BUILTINS_{target}_CMAKE_CXX_COMPILER_TARGET"] = target
            opts[f"BUILTINS_{target}_CMAKE_ASM_COMPILER_TARGET"] = target
            opts[f"BUILTINS_{target}_CMAKE_TRY_COMPILE_TARGET_TYPE"] = "STATIC_LIBRARY"
            opts[
                f"BUILTINS_{target}_CMAKE_EXE_LINKER_FLAGS"
            ] = f"--target={target} {ldflags}"
            opts[
                f"BUILTINS_{target}_CMAKE_SHARED_LINKER_FLAGS"
            ] = f"--target={target} {ldflags}"
            opts[f"BUILTINS_{target}_LLVM_USE_LINKER"] = "lld"
            opts[f"BUILTINS_{target}_COMPILER_RT_BAREMETAL_BUILD"] = "ON"
            opts[f"BUILTINS_{target}_COMPILER_RT_DEFAULT_TARGET_ONLY"] = "ON"

            # May be we need to build the builtins with the host compiler
            # "CMAKE_C_COMPILER_EXTERNAL_TOOLCHAIN": str(gcc_toolchain_path),
            # "CMAKE_CXX_COMPILER_EXTERNAL_TOOLCHAIN": str(gcc_toolchain_path),
            # "CMAKE_SYSROOT": str(sysroot_path),

            opts[f"RUNTIMES_{target}_CMAKE_SYSTEM_NAME"] = "Linux"
            opts[f"RUNTIMES_{target}_CMAKE_BUILD_TYPE"] = "Release"
            opts[f"RUNTIMES_{target}_CMAKE_C_FLAGS"] = f"--target={target}"
            opts[f"RUNTIMES_{target}_CMAKE_CXX_FLAGS"] = f"--target={target}"
            opts[f"RUNTIMES_{target}_CMAKE_ASM_FLAGS"] = f"--target={target}"
            opts[f"RUNTIMES_{target}_CMAKE_TRY_COMPILE_TARGET_TYPE"] = "STATIC_LIBRARY"
            opts[f"RUNTIMES_{target}_CMAKE_VERBOSE_MAKEFILE"] = "ON"
            opts[f"RUNTIMES_{target}_CMAKE_POSITION_INDEPENDENT_CODE"] = "ON"
            opts[f"RUNTIMES_{target}_LLVM_USE_LINKER"] = "lld"
            opts[f"RUNTIMES_{target}_COMPILER_RT_BUILD_SANITIZERS"] = "OFF"
            opts[f"RUNTIMES_{target}_COMPILER_RT_BUILD_XRAY"] = "OFF"
            opts[f"RUNTIMES_{target}_COMPILER_RT_BUILD_LIBFUZZER"] = "OFF"
            opts[f"RUNTIMES_{target}_COMPILER_RT_BUILD_MEMPROF"] = "OFF"
            opts[f"RUNTIMES_{target}_COMPILER_RT_BUILD_ORC"] = "OFF"
            opts[f"RUNTIMES_{target}_COMPILER_RT_BUILD_GWP_ASAN"] = "OFF"
            opts[f"RUNTIMES_{target}_COMPILER_RT_BUILD_PROFILE"] = "ON"
            opts[f"RUNTIMES_{target}_COMPILER_RT_USE_BUILTINS_LIBRARY"] = "ON"
            opts[f"RUNTIMES_{target}_LIBCXX_CXX_ABI"] = "libcxxabi"
            opts[f"RUNTIMES_{target}_LIBCXX_USE_COMPILER_RT"] = "ON"
            opts[f"RUNTIMES_{target}_LIBCXX_ENABLE_STATIC_ABI_LIBRARY"] = "ON"
            opts[f"RUNTIMES_{target}_LIBCXX_ABI_UNSTABLE"] = "ON"
            opts[
                f"RUNTIMES_{target}_LIBCXX_STATICALLY_LINK_ABI_IN_SHARED_LIBRARY"
            ] = "ON"
            opts[
                f"RUNTIMES_{target}_LIBCXX_STATICALLY_LINK_ABI_IN_STATIC_LIBRARY"
            ] = "ON"
            opts[f"RUNTIMES_{target}_LIBCXXABI_ENABLE_THREADS"] = "ON"
            opts[f"RUNTIMES_{target}_LIBCXXABI_HAS_CXA_THREAD_ATEXIT_IMPL"] = "OFF"
            opts[f"RUNTIMES_{target}_LIBCXXABI_USE_COMPILER_RT"] = "ON"
            opts[f"RUNTIMES_{target}_LIBCXXABI_USE_LLVM_UNWINDER"] = "ON"
            opts[f"RUNTIMES_{target}_LIBCXXABI_ENABLE_STATIC_UNWINDER"] = "ON"
            opts[f"RUNTIMES_{target}_LIBUNWIND_USE_COMPILER_RT"] = "ON"
            opts[f"RUNTIMES_{target}_LIBCXX_ENABLE_STATIC_ABI_LIBRARY"] = "ON"
            opts[f"RUNTIMES_{target}_LIBCXXABI_ENABLE_SHARED"] = "OFF"
            opts[f"RUNTIMES_{target}_LIBCXX_ABI_VERSION"] = "2"

        # opts[f"RUNTIMES_{target}_CMAKE_C_COMPILER_EXTERNAL_TOOLCHAIN"] = gcc_toolchain_path
        # opts[f"RUNTIMES_{target}_CMAKE_CXX_COMPILER_EXTERNAL_TOOLCHAIN"] = gcc_toolchain_path

        llvm_dir = self.dependencies["llvm-source"].buildenv_info.vars(
            self, scope="build"
        )

        # opts["CMAKE_CXX_FLAGS"] = f"{envs.get('CXXFLAGS', '')} -g -g3 -Wl,-fuse-ld=lld"
        # opts["CMAKE_C_FLAGS"] = f"{envs.get('CFLAGS', '')} -g -g3 -Wl,-fuse-ld=lld"
        # opts["CMAKE_EXE_LINKER_FLAGS"] = f"{envs.get('LDFLAGS', '')} -static-libgcc"
        # opts[
        #     "CMAKE_SHARED_LINKER_FLAGS"
        # ] = f"{envs.get('LDFLAGS', '')} -L{zlib_lib} -Bstatic -lz"

        build_dir = os.path.join(self.build_folder, "build")

        os.makedirs(build_dir)
        with current_dir(build_dir):
            cm = CMake(self)
            cm.configure(
                # We do not want `-static` with llvm eventhough we want to build
                # clan statically. That flag will ruin some of the shared libs that
                # has to be built for clang
                opts,
                build_script_folder=os.path.join(llvm_dir["LLVM_SOURCE_DIR"], "llvm"),
            )

            cm.build(target="install-distribution-stripped")
            # cm.build(target="install-runtimes-stripped")
            # cm.build(target="install-compiler-rt-stripped")
            # cm.build(target="install-cxx-stripped")
            # cm.build(target="install-cxxabi-stripped")
            shutil.copy(
                f"{self.build_folder}/bin/clang-tblgen",
                os.path.join(self.package_folder, "bin"),
            )

        shutil.rmtree("build")

    def package_info(self):
        bindir = Path(os.path.join(self.package_folder, "bin"))
        clang_lib_dir = Path(
            os.path.join(
                self.package_folder, "lib", "clang", get_version("llvm_major"), "lib"
            )
        )

        self.runenv_info.prepend_path(
            "PATH",
            str(bindir),
        )

        self.buildenv_info.define("CLANG_BOOTSTRAP_PATH", self.package_folder)
        self.buildenv_info.define("LIBCC_GNU", str(clang_lib_dir / self.gnu_triple))
        self.buildenv_info.define("LIBCC_MUSL", str(clang_lib_dir / self.musl_triple))
        self.buildenv_info.define("CC", str(bindir / "clang"))
        self.buildenv_info.define("CXX", str(bindir / "clang++"))
        self.buildenv_info.define("AR", str(bindir / "llvm-ar"))
        self.buildenv_info.define("NM", str(bindir / "llvm-nm"))
        self.buildenv_info.define("RANLIB", str(bindir / "llvm-ranlib"))

        self.buildenv_info.define("BOOTSTRAP_GNU_BUILD_TRIPLE", self.gnu_triple)
        self.runenv_info.define("BOOTSTRAP_GNU_BUILD_TRIPLE", self.gnu_triple)

        self.buildenv_info.define("BOOTSTRAP_MUSL_BUILD_TRIPLE", self.musl_triple)
        self.runenv_info.define("BOOTSTRAP_MUSL_BUILD_TRIPLE", self.musl_triple)
