# Toolchain builder for the Serene programming language
#
# Copyright (c) 2019-2023 Sameer Rahmani <lxsameer@gnu.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
from pathlib import Path

from conan import ConanFile
from conan.tools.cmake import CMake, CMakeToolchain, CMakeDeps

from conan.tools.files import get

from conf.utils import (
    with_static_flags,
    current_dir,
    get_version,
    copy_tree,
    copy_dependency_tree,
)

TOOLCHAIN_TARGETS = [
    "runtimes",
    "compiler-rt",
]

PROJECTS = []
RUNTIME_TARGETS = [
    "compiler-rt",
    "libcxx",
    "libcxxabi",
    "libunwind",
]

CMAKE_OPTIONS = {
    "LLVM_USE_LINKER": "lld",
    "LLVM_ENABLE_NEW_PASS_MANAGER": "ON",
    "LLVM_BUILD_TESTS": "OFF",
    "LLVM_ENABLE_ASSERTIONS": "OFF",
    "LLVM_ENABLE_LIBXML2": "OFF",
    "LLVM_ENABLE_TERMINFO": "OFF",
    "LLVM_STATIC_LINK_CXX_STDLIB": "ON",
    "LLVM_ENABLE_PER_TARGET_RUNTIME_DIR": "ON",
    "ENABLE_X86_RELAX_RELOCATIONS": "ON",
    "BUILD_SHARED_LIBS": "OFF",
    "LIBCXX_ENABLE_STATIC_ABI_LIBRARY": "ON",
    "LIBCXXABI_ENABLE_SHARED": "OFF",
    "LIBCXX_ABI_VERSION": "2",
    "LLVM_ENABLE_LTO": "THIN",
    "CMAKE_POSITION_INDEPENDENT_CODE": "ON",
    "LLVM_ENABLE_LIBCXX": "ON",
    "LLVM_THINLTO_CACHE_PATH": "/tmp/llvm-build-lto-libcxx",
    "LLVM_ENABLE_PROJECTS": ";".join(PROJECTS),
    "LIBCXX_HAS_MUSL_LIBC": "ON",
    "LLVM_DEFAULT_TARGET_TRIPLE": os.environ["TARGET"],
    "LIBCXX_CXX_ABI": "libcxxabi",
    "LIBCXX_USE_COMPILER_RT": "ON",
    "LIBCXX_ABI_UNSTABLE": "ON",
    "LIBCXX_STATICALLY_LINK_ABI_IN_SHARED_LIBRARY": "ON",
    "LIBCXX_STATICALLY_LINK_ABI_IN_STATIC_LIBRARY": "ON",
    "LIBCXX_HAS_ATOMIC_LIB": "OFF",
    "LIBCXXABI_ENABLE_THREADS": "ON",
    "LIBCXXABI_HAS_CXA_THREAD_ATEXIT_IMPL": "OFF",
    "LIBCXXABI_USE_COMPILER_RT": "ON",
    "LIBCXXABI_USE_LLVM_UNWINDER": "ON",
    "LIBCXXABI_ENABLE_STATIC_UNWINDER": "ON",
    "LIBCXXABI_STATICALLY_LINK_UNWINDER_IN_SHARED_LIBRARY": "ON",
    "LIBCXXABI_STATICALLY_LINK_UNWINDER_IN_STATIC_LIBRARY": "ON",
    "CMAKE_C_COMPILER_TARGET": os.environ["TARGET"],
    "CMAKE_CXX_COMPILER_TARGET": os.environ["TARGET"],
    "CMAKE_ASM_COMPILER_TARGET": os.environ["TARGET"],
    "CMAKE_INSTALL_RPATH": "\\$ORIGIN/../lib",
    "CMAKE_BUILD_WITH_INSTALL_RPATH": "ON",
    "COMPILER_RT_BAREMETAL_BUILD": "ON",
    "COMPILER_RT_DEFAULT_TARGET_ONLY": "ON",
    "CMAKE_VERBOSE_MAKEFILE": "ON",
    "COMPILER_RT_SANITIZERS_TO_BUILD": "asan;msan;tsan",
    "COMPILER_RT_USE_BUILTINS_LIBRARY": "ON",
    "COMPILER_RT_USE_LIBCXX": "ON",
    "COMPILER_RT_BUILD_SANITIZERS": "ON",
    "COMPILER_RT_HAS_GNU_VERSION_SCRIPT_COMPAT": "OFF",
    "COMPILER_RT_BUILD_XRAY": "OFF",
    "COMPILER_RT_BUILD_MEMPROF": "ON",
    "COMPILER_RT_BUILD_LIBFUZZER": "ON",
    "COMPILER_RT_BUILD_ORC": "ON",
    "SANITIZER_CXX_ABI": "libc++",
    "SANITIZER_CXX_ABI_INTREE": "ON",
    "LIBUNWIND_USE_COMPILER_RT": "ON",
}


class LibCXX(ConanFile):
    name = "libcxx"
    settings = "os", "arch", "compiler", "build_type"
    version = get_version("llvm")

    def build_requirements(self):
        self.requires(f"llvm-source/{self.version}@{self.user}/{self.channel}")

        self.requires(f"musl/{get_version('musl')}@{self.user}/{self.channel}")
        self.requires(f"zstd/{get_version('zstd')}@{self.user}/{self.channel}")
        self.requires(f"zlib/{get_version('zlib')}@{self.user}/{self.channel}")
        self.requires(
            f"clang-bootstrap/{get_version('llvm')}@{self.user}/{self.channel}"
        )
        self.tool_requires(f"cmake/{get_version('cmake')}@{self.user}/{self.channel}")
        self.tool_requires(f"ninja/{get_version('ninja')}@{self.user}/{self.channel}")

    @property
    def buildenv(self):
        return self.buildenv_info.vars(self, scope="build")

    def create_sysroot(self):
        sysroot = Path(self.build_folder) / "sysroot"
        sysroot.mkdir()

        copy_dependency_tree(self, ["clang-bootstrap", "musl", "zlib", "zstd"], sysroot)

    def add_compiler_options(self, opts):
        tc_dir = Path(self.build_folder) / "sysroot"
        opts["CMAKE_C_COMPILER"] = str(tc_dir / "bin" / "clang")
        opts["CMAKE_CXX_COMPILER"] = str(tc_dir / "bin" / "clang++")
        opts["LLVM_CONFIG_PATH"] = str(tc_dir / "bin" / "llvm-config")

        target = os.environ["TARGET"]

        opts["CMAKE_C_COMPILER_TARGET"] = target
        opts["CMAKE_CXX_COMPILER_TARGET"] = target
        opts["CMAKE_SYSROOT"] = str(tc_dir)

        opts["CMAKE_EXE_LINKER_FLAGS"] = f"-Wl,-rpath,{str(tc_dir)}/lib/{target}"
        opts["CMAKE_SHARED_LINKER_FLAGS"] = f"-Wl,-rpath,{str(tc_dir)}/lib/{target}"

    def add_runtimes_and_builtins(self, opts):
        builtin_targets = []
        target = os.environ["TARGET"]
        sysroot_path = Path(self.build_folder) / "sysroot"

        CMAKE_OPTIONS.update(
            {
                "CMAKE_SYSTEM_NAME": "Linux",
                "CMAKE_BUILD_TYPE": "Release",
                "CMAKE_SYSTEM_PROCESSOR": "X86_64",
                "CMAKE_C_FLAGS": self.buildenv.get("CFLAGS", ""),
                "CMAKE_CXX_FLAGS": self.buildenv.get("CXXFLAGS", ""),
                "CMAKE_ASM_FLAGS": self.buildenv.get("ASFLAGS", ""),
                "CMAKE_C_COMPILER_TARGET": target,
                "CMAKE_CXX_COMPILER_TARGET": target,
                "CMAKE_ASM_COMPILER_TARGET": target,
                "CMAKE_TRY_COMPILE_TARGET_TYPE": "STATIC_LIBRARY",
                "CMAKE_EXE_LINKER_FLAGS": f"--target={target} {self.buildenv.get('LDFLAGS', '')}",
                "CMAKE_SHARED_LINKER_FLAGS": f"--target={target} {self.buildenv.get('LDFLAGS', '')}",
                "CMAKE_SYSROOT": str(sysroot_path),
            }
        )

        opts["LLVM_RUNTIME_TARGETS"] = target
        opts["LLVM_ENABLE_RUNTIMES"] = ";".join(RUNTIME_TARGETS)

    def generate(self):
        tc = CMakeToolchain(self)
        tc.preprocessor_definitions["_LIBCPP_PROVIDES_DEFAULT_RUNE_TABLE"] = ""
        tc.preprocessor_definitions["_LIBCPP_HAS_MUSL_LIBC"] = "ON"

        tc.generate()
        deps = CMakeDeps(self)
        deps.set_property("zlib", "cmake_find_mode", "both")
        deps.generate()

    def build(self):
        self.create_sysroot()
        opts = CMAKE_OPTIONS.copy()
        sysroot = f"{self.build_folder}/sysroot"
        self.add_compiler_options(opts)

        opts[
            "CMAKE_REQUIRED_INCLUDES"
        ] = f"{sysroot}/lib/clang/{get_version('llvm_major')}/include {sysroot}/include"
        opts["CMAKE_REQUIRED_LIBRARIES"] = f"-L{sysroot}/lib -lz"

        self.add_runtimes_and_builtins(opts)

        for target in opts["LLVM_RUNTIME_TARGETS"].split(";"):
            extra_target = None

            bin_dir = Path(self.package_folder) / "bin"
            bin_dir.mkdir(exist_ok=True, parents=True)

        opts["LLVM_ENABLE_PROJECTS"] = ";".join(PROJECTS)
        opts["LLVM_DISTRIBUTION_COMPONENTS"] = ";".join(TOOLCHAIN_TARGETS)

        stage1 = Path(sysroot)

        llvm_dir = self.dependencies["llvm-source"].buildenv_info.vars(
            self, scope="build"
        )

        build_dir = os.path.join(self.build_folder, "build")
        os.makedirs(build_dir)

        with current_dir(build_dir):
            cm = CMake(self)
            cm.configure(
                opts,
                build_script_folder=os.path.join(
                    llvm_dir["LLVM_SOURCE_DIR"], "runtimes"
                ),
            )
            cm.build(target="cxx")
            cm.build(target="cxxabi")
            cm.build(target="unwind")
            cm.build(target="install-cxx")
            cm.build(target="install-cxxabi")
            cm.build(target="install-unwind")

    def package_info(self):
        self.buildenv_info.prepend_path(
            "LD_LIBRARY_PATH", os.path.join(self.package_folder, "lib")
        )
