# Toolchain builder for the Serene programming language
#
# Copyright (c) 2019-2023 Sameer Rahmani <lxsameer@gnu.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
from pathlib import Path

from conan import ConanFile
from conan.tools.files import get
from conan.tools.cmake import CMakeToolchain

from conf.utils import get_version, copy_template, copy_tree, copy_dependency_tree


class Stage2Sysroot(ConanFile):
    name = "stage2-sysroot"
    version = get_version("llvm")
    settings = "os", "arch"
    exports = (
        "*.cmake",
        "sysroot/*",
    )

    def build(self):
        # sysroot dir is created via the builder script
        copy_tree(
            f"{self.recipe_folder}/sysroot/{get_version('llvm')}/bin",
            f"{self.package_folder}/bin",
        )
        copy_tree(
            f"{self.recipe_folder}/sysroot/{get_version('llvm')}/lib",
            f"{self.package_folder}/lib",
        )
        copy_tree(
            f"{self.recipe_folder}/sysroot/{get_version('llvm')}/include",
            f"{self.package_folder}/include",
        )

        toolchain = f"{self.recipe_folder}/toolchain.cmake"
        copy_template(
            self,
            toolchain,
            f"{self.package_folder}/toolchain.cmake",
            {"TRIPLE": os.environ["TARGET"], "SYSROOT": self.package_folder},
        )

    def package_info(self):
        bindir = Path(self.package_folder) / "bin"
        self.runenv_info.prepend_path(
            "PATH",
            str(bindir),
        )

        self.buildenv_info.append("sysroot", self.package_folder)

        f = os.path.join(self.package_folder, "toolchain.cmake")
        self.conf_info.prepend("tools.cmake.cmaketoolchain:user_toolchain", f)
        self.conf_info.define("tools.build:sysroot", self.package_folder)
        self.buildenv_info.define_path("SYSROOT", self.package_folder)

        self.buildenv_info.define("CC", str(bindir / "clang"))
        self.buildenv_info.define("CXX", str(bindir / "clang++"))
        self.buildenv_info.define("AR", str(bindir / "clang"))
        self.buildenv_info.define("NM", str(bindir / "llvm-nm"))
        self.buildenv_info.define("RANLIB", str(bindir / "llvm-ranlib"))

        self.cpp_info.includedirs = [
            f"{self.package_folder}/include/c++/v1",
            f"{self.package_folder}/include/",
            f"{self.package_folder}/lib/clang/{get_version('llvm_major')}/include",
        ]
