# Toolchain builder for the Serene programming language
#
# Copyright (c) 2019-2023 Sameer Rahmani <lxsameer@gnu.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import shutil
from pathlib import Path

from conan import ConanFile
from conan.tools.cmake import CMake, CMakeToolchain, CMakeDeps

from conan.tools.files import get

from conf.utils import with_static_flags, current_dir, get_version

TOOLCHAIN_TARGETS = [
    "runtimes",
    "clang-apply-replacements",
    "clang-format",
    "clang-query",
    "clang-resource-headers",
    "clang-tidy",
    "clang-doc",
    "clang",
    "clangd",
    "clang-extdef-mapping",
    "cmake-exports",
    "dsymutil",
    "lld",
    "llvm-addr2line",
    "llvm-ar",
    "llvm-tblgen",
    "llvm-config",
    "llvm-as",
    "clang-tblgen",
    "llvm-cov",
    "llvm-cvtres",
    "llvm-cxxmap",
    "llvm-dlltool",
    "llvm-dwp",
    "llvm-dwarfdump",
    "llvm-install-name-tool",
    "llvm-lib",
    "llvm-lipo",
    "llvm-nm",
    "llvm-objcopy",
    "llvm-objdump",
    "llvm-pdbutil",
    "llvm-profdata",
    "llvm-ranlib",
    "llvm-rc",
    "llvm-readelf",
    "llvm-strings",
    "llvm-strip",
    "llvm-symbolizer",
    "llvm-windres",
    "LTO",
    "builtins",
    "llvm-libraries",
    "mlir-libraries",
    "mlir-tblgen",
    "mlir-opt",
    "mlir-lsp-server",
    "mlir-reduce",
    "llvm-jitlink",
    "compiler-rt",
    "runtimes",
]

# No lldb? Yes, lldb links against liblldb.so and since we use
# -static with our executables that wouldn't be easy to handle
# and I'm lazy. We can just compile the lldb or any other debugger
# on the host
PROJECTS = ["clang-tools-extra", "clang", "lld", "llvm", "mlir"]
RUNTIME_TARGETS = [
    "compiler-rt",
    "libcxx",
    "libcxxabi",
    "libunwind",
]

CMAKE_OPTIONS = {
    "LLVM_USE_LINKER": "lld",
    "LLVM_ENABLE_NEW_PASS_MANAGER": "ON",
    "LLVM_BUILD_TESTS": "OFF",
    "LLVM_ENABLE_ASSERTIONS": "ON",
    "LLVM_ENABLE_LIBXML2": "OFF",
    "LLVM_ENABLE_TERMINFO": "OFF",
    "LLVM_ENABLE_ZLIB": "FORCE_ON",
    "LLVM_INCLUDE_BENCHMARKS": "OFF",
    "LLVM_INCLUDE_EXAMPLES": "OFF",
    "LLVM_INCLUDE_TESTS": "OFF",
    "LLVM_INCLUDE_GO_TESTS": "OFF",
    "LLVM_ENABLE_BINDINGS": "OFF",
    "LLVM_TARGETS_TO_BUILD": "X86;AArch64;AMDGPU;ARM;RISCV;WebAssembly",
    "LLVM_STATIC_LINK_CXX_STDLIB": "ON",
    "PACKAGE_VENDOR": "Serene",
    "LLVM_ENABLE_PER_TARGET_RUNTIME_DIR": "ON",
    "ENABLE_X86_RELAX_RELOCATIONS": "ON",
    "BUILD_SHARED_LIBS": "OFF",
    "CLANG_ENABLE_BOOTSTRAP": "ON",
    "LIBCXX_ENABLE_STATIC_ABI_LIBRARY": "ON",
    "LIBCXXABI_ENABLE_SHARED": "OFF",
    "LIBCXX_ABI_VERSION": "2",
    "LLVM_CCACHE_BUILD": "ON",
    # TODO: At this point lld crashes during linking clang itself
    # cause of an issues in the `MachineDominator Tree Construction`
    # pass that leads to an out of memory issue. Turn LTO on later
    # "LLVM_ENABLE_LTO": "THIN",
    "CMAKE_POSITION_INDEPENDENT_CODE": "ON",
    "CLANG_DEFAULT_CXX_STDLIB": "libc++",
    "CLANG_DEFAULT_LINKER": "lld",
    "CLANG_DEFAULT_OBJCOPY": "llvm-objcopy",
    "CLANG_DEFAULT_RTLIB": "compiler-rt",
    "CLANG_VENDOR_UTI": "serene.toolchain",
    "LLVM_ENABLE_LIBCXX": "ON",
    "LLVM_ENABLE_ZSTD": "OFF",
    "LLVM_OPTIMIZED_TABLEGEN": "OFF",
    "LLVM_THINLTO_CACHE_PATH": "/tmp/llvm-build-lto",
    "LLVM_ENABLE_PROJECTS": ";".join(PROJECTS),
    "LLVM_CCACHE_BUILD": "ON",
    "LIBCXX_HAS_MUSL_LIBC": "ON",
    # "LLVM_PARALLEL_COMPILE_JOBS": "32",
    # "LLVM_PARALLEL_LINK_JOBS": "1",
    "LLVM_DEFAULT_TARGET_TRIPLE": os.environ["TARGET"],
    # Common
    "CMAKE_POSITION_INDEPENDENT_CODE": "ON",
}


class toolchain(ConanFile):
    name = "toolchain"
    settings = "os", "arch", "build_type"
    version = get_version("llvm")

    def build_requirements(self):
        self.requires(f"llvm-source/{self.version}@{self.user}/{self.channel}")

        self.requires(
            f"stage2-sysroot/{get_version('llvm')}@{self.user}/{self.channel}"
        )

    @property
    def buildenv(self):
        return self.buildenv_info.vars(self, scope="build")

    def add_compiler_options(self, opts):
        # We can't use the clang-bootstrap toolchain file here because it will be passed
        # to the runtimes and bulitins subcmake stuff later and it will "poison" the cmake
        # cache with the wrong compiler and options. So we do this manually the hard way.

        tc_dir = Path(self.dependencies["stage2-sysroot"].package_folder)
        opts["CMAKE_C_COMPILER"] = str(tc_dir / "bin" / "clang")
        opts["CMAKE_CXX_COMPILER"] = str(tc_dir / "bin" / "clang++")
        # opts["LLVM_CONFIG_PATH"] = str(tc_dir / "bin" / "llvm-config")

        target = os.environ["TARGET"]

        opts["CMAKE_C_COMPILER_TARGET"] = target
        opts["CMAKE_CXX_COMPILER_TARGET"] = target
        opts["CMAKE_SYSROOT"] = str(tc_dir)

        opts[
            "CMAKE_EXE_LINKER_FLAGS"
        ] = f"-Wl,-rpath,{str(tc_dir)}/lib/{target} -static -Wl,-static"
        opts[
            "CMAKE_SHARED_LINKER_FLAGS"
        ] = f"-Wl,-rpath,$ORIGIN/../lib/{target} -Wl,-rpath,$ORIGIN/../lib/"
        # opts["CMAKE_C_FLAGS"] = "-static" -Wl,-rpath,{str(tc_dir)}/lib/{target}
        # opts["CMAKE_CXX_FLAGS"] = "-static"

    def add_runtimes_and_builtins(self, opts):
        builtin_targets = []
        target = os.environ["TARGET"]
        sysroot_path = Path(self.dependencies["stage2-sysroot"].package_folder)

        target_builtin_opts = {
            "CMAKE_SYSTEM_NAME": "Linux",
            "CMAKE_BUILD_TYPE": self.settings.build_type,
            "CMAKE_SYSTEM_PROCESSOR": "X86_64",
            "CMAKE_C_FLAGS": self.buildenv.get("CFLAGS", ""),
            "CMAKE_CXX_FLAGS": self.buildenv.get("CXXFLAGS", ""),
            "CMAKE_ASM_FLAGS": self.buildenv.get("ASFLAGS", ""),
            "CMAKE_C_COMPILER_TARGET": target,
            "CMAKE_CXX_COMPILER_TARGET": target,
            "CMAKE_ASM_COMPILER_TARGET": target,
            "CMAKE_TRY_COMPILE_TARGET_TYPE": "STATIC_LIBRARY",
            "CMAKE_EXE_LINKER_FLAGS": f"--target={target} {self.buildenv.get('LDFLAGS', '')}",
            "CMAKE_SHARED_LINKER_FLAGS": f"--target={target} {self.buildenv.get('LDFLAGS', '')}",
            "CMAKE_SYSROOT": str(sysroot_path),
            "CMAKE_INSTALL_RPATH": "\$ORIGIN/../lib",
            "CMAKE_BUILD_WITH_INSTALL_RPATH": "ON",
            "LLVM_USE_LINKER": "lld",
            "COMPILER_RT_BAREMETAL_BUILD": "ON",
            "COMPILER_RT_DEFAULT_TARGET_ONLY": "ON",
            "LLVM_CCACHE_BUILD": opts["LLVM_CCACHE_BUILD"],
        }

        for k, v in target_builtin_opts.items():
            opts[f"BUILTINS_{target}_{k}"] = v

        target_runtime_opts = {
            "CMAKE_SYSTEM_NAME": "Linux",
            "CMAKE_BUILD_TYPE": self.settings.build_type,
            "CMAKE_C_FLAGS": self.buildenv.get("CFLAGS", ""),
            "CMAKE_CXX_FLAGS": self.buildenv.get("CXXFLAGS", ""),
            "CMAKE_ASM_FLAGS": self.buildenv.get("ASFLAGS", ""),
            "CMAKE_C_COMPILER_TARGET": target,
            "CMAKE_CXX_COMPILER_TARGET": target,
            "CMAKE_ASM_COMPILER_TARGET": target,
            "CMAKE_TRY_COMPILE_TARGET_TYPE": "STATIC_LIBRARY",
            "CMAKE_EXE_LINKER_FLAGS": f"--target={target} {self.buildenv.get('LDFLAGS', '')}",
            "CMAKE_SHARED_LINKER_FLAGS": f"--target={target} {self.buildenv.get('LDFLAGS', '')}",
            "CMAKE_INSTALL_RPATH": "\$ORIGIN/../lib",
            "CMAKE_BUILD_WITH_INSTALL_RPATH": "ON",
            "CMAKE_SYSROOT": str(sysroot_path),
            "CMAKE_VERBOSE_MAKEFILE": "ON",
            "CMAKE_POSITION_INDEPENDENT_CODE": "ON",
            "LLVM_USE_LINKER": "lld",
            "LLVM_USE_LTO": "ON",
            "LLVM_CCACHE_BUILD": opts["LLVM_CCACHE_BUILD"],
            # TODO: Check for more sanitizers that work with musl
            "COMPILER_RT_SANITIZERS_TO_BUILD": "asan;msan;tsan",
            "COMPILER_RT_USE_BUILTINS_LIBRARY": "ON",
            "COMPILER_RT_USE_LIBCXX": "ON",
            "COMPILER_RT_BUILD_SANITIZERS": "ON",
            # For some reason this flag is not correctly set if we don't
            # manually set it and some strange gnu only LDFLAGS are injected.
            "COMPILER_RT_HAS_GNU_VERSION_SCRIPT_COMPAT": "OFF",
            # X-Ray doesn't seem to compiler with musl
            "COMPILER_RT_BUILD_XRAY": "OFF",
            # Only build these if we enable sanitizers since they depend
            # on the sanitizer common runtime
            "COMPILER_RT_BUILD_MEMPROF": "ON",
            "COMPILER_RT_BUILD_LIBFUZZER": "ON",
            "COMPILER_RT_BUILD_ORC": "ON",
            # Make sure we use libc++ from the tree instead from the system.
            "SANITIZER_CXX_ABI": "libc++",
            "SANITIZER_CXX_ABI_INTREE": "ON",
            "LIBCXX_CXX_ABI": "libcxxabi",
            "LIBCXX_USE_COMPILER_RT": "ON",
            "LIBCXX_ENABLE_STATIC_ABI_LIBRARY": "ON",
            "LIBCXX_ABI_UNSTABLE": "ON",
            "LIBCXX_STATICALLY_LINK_ABI_IN_SHARED_LIBRARY": "ON",
            "LIBCXX_STATICALLY_LINK_ABI_IN_STATIC_LIBRARY": "ON",
            "LIBCXX_ABI_VERSION": "2",
            "LIBCXX_HAS_MUSL_LIBC": "ON",
            # musl doesn't have -latomic
            "LIBCXX_HAS_ATOMIC_LIB": "OFF",
            "LIBCXXABI_ENABLE_THREADS": "ON",
            "LIBCXXABI_HAS_CXA_THREAD_ATEXIT_IMPL": "OFF",
            "LIBCXXABI_USE_COMPILER_RT": "ON",
            "LIBCXXABI_USE_LLVM_UNWINDER": "ON",
            "LIBCXXABI_ENABLE_STATIC_UNWINDER": "ON",
            "LIBCXXABI_STATICALLY_LINK_UNWINDER_IN_SHARED_LIBRARY": "ON",
            "LIBCXXABI_STATICALLY_LINK_UNWINDER_IN_STATIC_LIBRARY": "ON",
            "LIBCXXABI_ENABLE_SHARED": "OFF",
            "LIBUNWIND_USE_COMPILER_RT": "ON",
        }

        for k, v in target_runtime_opts.items():
            opts[f"RUNTIMES_{target}_{k}"] = v

        opts["LLVM_RUNTIME_TARGETS"] = target
        opts["LLVM_BUILTIN_TARGETS"] = target
        opts["LLVM_ENABLE_RUNTIMES"] = ";".join(RUNTIME_TARGETS)

    def generate(self):
        tc = CMakeToolchain(self)
        tc.preprocessor_definitions["_LIBCPP_PROVIDES_DEFAULT_RUNE_TABLE"] = ""
        tc.preprocessor_definitions["_LIBCPP_HAS_MUSL_LIBC"] = "ON"

        tc.generate()
        deps = CMakeDeps(self)
        deps.set_property("zlib", "cmake_find_mode", "both")
        deps.generate()

    def build(self):
        opts = CMAKE_OPTIONS.copy()
        sysroot = self.dependencies["stage2-sysroot"].package_folder
        self.add_compiler_options(opts)

        # zlib_dir = self.dependencies["zlib"].package_folder
        opts["ZLIB_ROOT"] = sysroot
        opts[
            "CMAKE_REQUIRED_INCLUDES"
        ] = f"{sysroot}/lib/clang/17/include {sysroot}/include"
        opts["CMAKE_REQUIRED_LIBRARIES"] = f"-L{sysroot}/lib -lz"

        # opts["LLVM_EXTERNALIZE_DEBUGINFO"] = "ON"
        # opts["LLVM_DEBUGINFO_OUTPUT_DIR"] = f"{self.package_folder}/debug"

        self.add_runtimes_and_builtins(opts)

        for target in opts["LLVM_RUNTIME_TARGETS"].split(";"):
            extra_target = None

            bin_dir = Path(self.package_folder) / "bin"
            bin_dir.mkdir(exist_ok=True, parents=True)

        opts["LLVM_ENABLE_PROJECTS"] = ";".join(PROJECTS)
        opts["LLVM_DISTRIBUTION_COMPONENTS"] = ";".join(TOOLCHAIN_TARGETS)

        stage1 = Path(self.dependencies["stage2-sysroot"].package_folder)
        opts["LLVM_TABLEGEN"] = str(stage1 / "bin" / "llvm-tblgen")
        opts["CLANG_TABLEGEN"] = str(stage1 / "bin" / "clang-tblgen")

        llvm_dir = self.dependencies["llvm-source"].buildenv_info.vars(
            self, scope="build"
        )

        build_dir = os.path.join(self.build_folder, "build")
        os.makedirs(build_dir)

        with current_dir(build_dir):
            cm = CMake(self)
            cm.configure(
                opts,
                build_script_folder=os.path.join(llvm_dir["LLVM_SOURCE_DIR"], "llvm"),
            )

            cm.build(target="install-builtins")
            cm.build(target="install-compiler-rt")
            cm.build(target="install-runtimes")
            # cm.build(target="install-toolchain-stripped")
            cm.build(target="install-distribution-stripped")

    def package_info(self):
        bindir = Path(os.path.join(self.package_folder, "bin"))
        clang_lib_dir = Path(
            os.path.join(
                self.package_folder, "lib", "clang", get_version("llvm_major"), "lib"
            )
        )

        self.runenv_info.prepend_path(
            "PATH",
            str(bindir),
        )

        self.buildenv_info.define("CC", str(bindir / "clang"))
        self.buildenv_info.define("CXX", str(bindir / "clang++"))
        self.buildenv_info.define("AR", str(bindir / "llvm-ar"))
        self.buildenv_info.define("NM", str(bindir / "llvm-nm"))
        self.buildenv_info.define("RANLIB", str(bindir / "llvm-ranlib"))
        self.buildenv_info.append("sysroot", self.package_folder)

        f = os.path.join(self.package_folder, "toolchain.cmake")
        self.conf_info.prepend("tools.cmake.cmaketoolchain:user_toolchain", f)
        self.conf_info.define("tools.build:sysroot", self.package_folder)
        self.buildenv_info.define_path("SYSROOT", self.package_folder)
